/*********************************************************************/
/* file: path.c - stuff for the path feature                         */
/*                             HUT TINTIN                            */
/*                coded by Ossi Tulonen from TinTin III              */
/*                TinTin III coded by peter unold 1992               */
/*********************************************************************/
/* the path is implemented as a fix-sized queue. It gets a bit messy */
/* here and there, but it should work....                            */
/*********************************************************************/
/*
 * TinTin III code Copyright (c) 1992 Peter Unold
 * All changed code from TinTin III Copyright (c) 1992-1995 Ossi tulonen
 * All rights reserved.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 */
 
#include "structs.h"

#include <string.h>
#include <ctype.h>
#include <assert.h>

#include "comms_help.h"
#include "parse.h"
#include "path.h"
#include "session.h"
#include "screen.h"
#include "utility.h"
#include "utils.h"

extern struct master_table_data master_table;
extern unsigned char tintin_char;
extern unsigned char text_start, text_end;

/*********************/
/* the #MARK command */
/*********************/
session *mark_command(const unsigned char *, session *ses)
{
  ses->reset_path();
  ses->print((const uchar *)"OK. MARK SET.\n\r");
  return(ses);
}

/*********************/
/* the #PATH command */
/*********************/

session *comm_path(const unsigned char *args, session *ses)
{
  uchar buf[BUFFER_SIZE], buf2[BUFFER_SIZE];
  unsigned char mypath[MAX_PATH_LENGTH + 30] = "#PATH YOU WENT: ";
  int len, pos;
  
  args = get_arg_with_spaces(args, buf);
  
  if(!*buf) {
    len = strlen((const char *)mypath);
    pos = ses->path_to_string(mypath + len, sizeof(mypath) - len);
    pos = str_cat(mypath, pos + len, sizeof(mypath), (const uchar *)"\n\r");
    ses->print(mypath);
    
  } else if(!check_insert_path(buf, ses)) {
    ses->print(help_path);
    
  } else if(IS_SET(ses->message_flags, MESSAGE_PATH)) {
    sprintf((char *)buf2, "# %s added to path.\n\r", buf);
    ses->print(buf2);
  }
  
  return(ses);
}

/* **********************
  the #savepath command
********************** */

session *comm_savepath(const unsigned char *arg, session *ses)
{
  unsigned char buf[BUFFER_SIZE], buf2[BUFFER_SIZE], path[MAX_PATH_LENGTH + 1];
  int pos;
  
  arg = get_arg_stop_spaces(arg, buf2);
  if(!*buf2) {
    ses->print(help_savepath);
    return(ses);
  }
  pos = ses->path_to_string(path, sizeof(path));
  
  if(!*path) {
    ses->print((const uchar *)"#ERROR - no path to put alias.\n\r");
    return(ses);
  }
  
  sprintf((char *)buf, "%calias %c%s%c %c%s%c", tintin_char,
	  text_start, buf2, text_end,
	  text_start, path, text_end);
  
  ses = parse_input(buf, ses);
  return(ses);
}

/***********************/
/* the #RETURN command */
/***********************/
session *return_command(const unsigned char *arg, session *ses)
{
  uchar ret_val = 0, buf[BUFFER_SIZE];
  unsigned char command[2]={'x', '\n'};
  int amount = 1;
  
  arg = get_arg_with_spaces(arg, buf);
  if(*buf) {
    if(stricmp(buf, (const uchar *)"all") == 0) {
      amount = -1;
    } else {
      amount = string_to_number(buf, ses);
    }
  }

  while((amount > 0 || amount == -1) && (ret_val = ses->remove_path())) {
    command[0] = reverse_dir(ret_val);
    
    if(ses->logfile)
      fwrite(command, 1, 2, ses->logfile);
    
    if(ses->get_socket() >= 0)
      if(write(ses->get_socket(), command, 2) == -1)
	ses->print_errmsg("write in send_to_mud");
    if(amount > 0)
      amount--;
  }
  return(ses);
}

/***********************/
/* the #UNPATH command */
/***********************/
session *unpath_command(const unsigned char *, session *ses)
{
  if(ses->remove_path() && IS_SET(ses->message_flags, MESSAGE_PATH)) {
    ses->print((const uchar *)"OK. FORGOT THAT MOVE.\n\r");
  }
  
  return(ses);
}

/*********************************************************/
/* check to see if command=direction then insert in path */
/*********************************************************/
int check_insert_path(const uchar *command, session *ses)
{
  if(is_abrev((const uchar *)"north", command)) {
    ses->insert_path('n');

  } else if(is_abrev((const uchar *)"east", command)) {
    ses->insert_path('e');

  } else if(is_abrev((const uchar *)"south", command)) {
    ses->insert_path('s');

  } else if(is_abrev((const uchar *)"west", command)) {
    ses->insert_path('w');

  } else if(is_abrev((const uchar *)"up", command)) {
    ses->insert_path('u');

  } else if(is_abrev((const uchar *)"down", command)) {
    ses->insert_path('d');
    
  } else {
    return(FALSE);
  }
  
  return(TRUE);
}

// *****************************************
//  code to insert direction in path buffer
// *****************************************

void session::insert_path(unsigned char dir)
{
  char buf[20];
  int len, i;
  
  if(path_lastdir == dir) {
    path_lastnum++;
  } else {
    if(path_lastnum > 0) {
      len = 0;
      if(path_lastnum > 1) {
	sprintf(buf, "%d", path_lastnum);
	len = strlen(buf);
      }
      buf[len++] = path_lastdir;
      buf[len]   = '\0';
      
      for(i = 0; i < len; i++) {
	if(path_length < MAX_PATH_LENGTH) {
	  path[(path_mark + path_length++)%MAX_PATH_LENGTH] = buf[i];
	} else {
	  if(isdigit(path[path_mark])) {
	    path[path_mark++] = buf[i];
	    if(path_mark == MAX_PATH_LENGTH)
	      path_mark = 0;
	    while(isdigit(path[path_mark])) {
	      path_mark++;
	      path_length--;
	      if(path_mark == MAX_PATH_LENGTH)
		path_mark = 0;
	    }
	    path_mark++;
	    path_length--;
	  } else {
	    path[path_mark++] = buf[i];
	  }
	  if(path_mark == MAX_PATH_LENGTH)
	    path_mark = 0;
	}
      }
    }
    path_lastdir = dir;
    path_lastnum = 1;
  }
}

// ************************************************
//  code to remove last direction from path buffer
// ************************************************

const uchar session::remove_path()
{
  char buf[30], *str;
  uchar ret_value;
  int i;
  
  if(path_lastnum > 0) {
    ret_value = path_lastdir;
    path_lastnum--;
    
    if(path_lastnum == 0 && path_length > 0) {
      path_length--;
      path_lastdir = path[(path_mark + path_length) % MAX_PATH_LENGTH];
      
      if(path_length > 0 &&
	 isdigit(path[(path_mark + path_length - 1) % MAX_PATH_LENGTH])) {
	str = buf;
	
	while(path_length > 0 &&
	      isdigit(path[(path_mark + path_length - 1) % MAX_PATH_LENGTH]))
	  path_length--;
	
	for(i = 0;
	    isdigit(path[(path_mark + path_length + i) % MAX_PATH_LENGTH]) &&
	    i < 30;
	    i++)
	  *str++ = path[(path_mark + path_length + i) % MAX_PATH_LENGTH];
	*str = '\0';
	
	path_lastnum = atoi(buf);
      } else {
	path_lastnum = 1;
      }
    }
    return(ret_value);
  }
  return(0);
}

// *************************************
//  code to make ascii string from path
// *************************************

const int session::path_to_string(uchar *buf, const int maxlen)
{
  uchar tmp[30];
  int pos = 0;
  
  assert(maxlen > MAX_PATH_LENGTH);
  
  if(path_lastnum > 0) {
    for(pos = 0; pos < path_length; pos++)
      buf[pos] = path[(path_mark + pos)%MAX_PATH_LENGTH];
    
    if(path_lastnum > 1) {
      sprintf((char *)tmp, "%d", path_lastnum);
      pos = str_cat(buf, pos, maxlen - 1, tmp);
    }
    buf[pos++] = path_lastdir;
  }
  buf[pos] = '\0';
  
  return(pos);
}

/***************************************************/
/* return char containing reverse direction as dir */
/***************************************************/
unsigned char reverse_dir(unsigned char dir)
{
  switch(dir) {
  case 'n':
    return 's';
  case 'e':
    return 'w';
  case 's':
    return 'n';
  case 'w':
    return 'e';
  case 'u':
    return 'd';
  case 'd':
    return 'u';
  }
  return 'x';
}
