#include "config.h" // here should be most config information about system
                    // done by configure script.

typedef unsigned char uchar;
#ifndef HAVE_ULONG
typedef unsigned long ulong;
#endif

#include <stdio.h>
#ifdef HAVE_STDLIB_H
#  include <stdlib.h>
#endif
#ifdef HAVE_UNISTD_H
#  include <unistd.h>
#endif
#include <sys/types.h>

extern "C" { // for freebsd 2 which doesn't know how to do it itself
#define _BSD_SIGNALS 1 // this is mainly for irix's but I hope it won't harm
                       // in other systems *stare*
#include <signal.h>
}

#if !defined(SA_RESTART) && defined(HAVE_SIGACTION) // lets use old signal func
#undef HAVE_SIGACTION                               // if no SA_RESTART
#endif

#ifdef HAVE_TERMIOS_H
#  include <termios.h>
#  include <sys/ioctl.h>
#else
#  include <sys/ioctl.h>
#  ifdef DIRTY_REDEFINE
#    undef TIOCGETP
#    define TIOCGETP 0x40067408
#  endif
#  include <sgtty.h>
#endif

#ifdef HAVE_LIBTERMCAP
extern "C" {
#ifdef HAVE_TERMCAP_H
#  include <termcap.h>
#else
  int tgetent(void *, const char *);
  int tgetnum(const char *);
  char *tgetstr(const char *, char **);
  char *tgoto(const char *, int, int);
#endif
}
#endif

#ifdef __HT_NET_C__
#define TELCMDS 1
#endif

#include <arpa/telnet.h>

#ifdef __HT_NET_C__
#undef TELCMDS
#define TELOPTS 1
#endif
#include "telnet.h"

#ifndef X_DISPLAY_MISSING
#  include <X11/Xlib.h>
#  include <X11/Xutil.h>
#  include <X11/Intrinsic.h>

#ifndef __HT_MAIN_C__
extern Display *MainDisplay;
#endif
#endif

/***********************************************/
/* Some default values you might wanna change: */
/***********************************************/

#define DEFAULT_TINTIN_CHAR '#'       /* tintin char                     */
#define DEFAULT_TEX_START_CHAR '['    /* start text field                */
#define DEFAULT_TEX_END_CHAR ']'      /* end text field                  */
#define DEFAULT_ECHO 0                /* echo                            */
#define DEFAULT_MESS MESSAGE_ALL
#define DEFAULT_IGNORE FALSE          /* ignore                          */
#define DEFAULT_CHARS_PER_LINE 80     /* chars per line if WinSize fail  */
#define DEFAULT_LINES_IN_SCR   24     /* lines in screen if WinSize fail */
#define DEFAULT_SES_HIST_SIZE  500    /* number of lines keep in memory  */
#define DEFAULT_RETURN_CHAR ';'       /* return char in command lines    */
#define DEFAULT_VERBATIM_CHAR '\\'    /* verbatim char                   */
#define DEFAULT_SES_FLAGS (SES_FLAG_EMAILCHECK | SES_FLAG_KEYBELL     | \
			   SES_FLAG_LINEMODE   | SES_FLAG_SPEEDWALK   | \
                           SES_FLAG_TERMECHO   | SES_FLAG_WFESPACE    | \
			   SES_FLAG_ASK_CONFIRM)
#define DEFAULT_MASTER_FLAGS (MASTER_FLAG_BELL | MASTER_FLAG_VISUAL_BELL)

#define HT_ERROR_FILE "tintin_error.log"

/**************************************************************************/
/* The stuff below here shouldn't be modified unless you know what you're */
/* doing........                                                          */
/**************************************************************************/

#ifndef TRUE
  #define TRUE 1
#endif

#ifndef FALSE
  #define FALSE 0
#endif

#define DEF_SHELL_ENV "SHELL"
#define DEF_TERM_TYPE (const uchar *)"vt100"
#define DEF_TIME_LINE (const uchar *)"Clock: %d %t%f%b%a(%w:%s) Time: %o"

#define BUFFER_SIZE 4096

#define HISTORY_SIZE 100              /* history size            */
#define MAX_PATH_LENGTH 100           /* max path length         */
#define MAX_PROMPT_SIZE 512           /* input prompt max length */
#define MAX_ALIAS_CALLS 10            /* max allowed alias calls */
                                      /* just to check forever loops */
#define MAX_REPEATIF_CALLS 10000      // max allowed repeatif commands
				      // on ONE line.
				      // just to check forever loops
#define MAX_WHILE_LOOPS 10000         // max allowed loops in while command
                                      // just to check forever loops
#define MAX_LINE_LENGTH 1024

// time in which need to be mouse button pressed again that it is
// considered double/trible pressing.
#define STYLE_MAX_TIME 300            // milliseconds


#define ESC 27

/* TIMELINEON  should be ansi codes for what kind text is 
   wanted time line to be ...                                      */
#define TIMELINEON  (const uchar *)"\033[7m"    /* reverse video             */

#define ECHO_ACTION 0x1
#define ECHO_IF     0x2
#define ECHO_CMP    0x4
#define ECHO_WAIT   0x8
#define ECHO_ALIAS  0x10

#define IGNORE_ACTION 0x1
#define IGNORE_SUBS   0x2
#define IGNORE_ALIAS  0x4
#define IGNORE_VAR    0x8
#define IGNORE_CALC   0x10
#define IGNORE_ALL    0x20

#define MESSAGE_ALL    (~0L)
#define MESSAGE_ACTION (1L<<0)
#define MESSAGE_ALIAS  (1L<<1)
#define MESSAGE_MACRO  (1L<<2)
#define MESSAGE_PATH   (1L<<3)
#define MESSAGE_SET    (1L<<4)
#define MESSAGE_SUBS   (1L<<5)
#define MESSAGE_VAR    (1L<<6)
#define MESSAGE_WAIT   (1L<<7)
#define MESSAGE_WORD   (1L<<8)

#define NULL_WIN (window *)-1

/************************ structures *********************/

class window;
class session;

class hutchar {
public:
  hutchar();
  ~hutchar();
  
  void clear();
  int is_middle() const { return(middle_pos ? TRUE : FALSE); }
  
  uchar *text, *middle_pos;
  int len;
};

class txt_block
{
public:
  txt_block(const uchar *);
  ~txt_block();
  
  uchar *text;
  
  txt_block *next;
};

class txt_q
{
public:
  txt_q();
  ~txt_q();
  
  void add_line(const uchar *);
  void get_line(uchar *, int);
  
  hutchar middle_txt;
  
private:
  struct txt_block *head, *tail;
};

class stack {  // code is in utility.cc
public:
  stack();
  ~stack();

  static stack *get_all() { return main; }
  stack *get_next()       { return next; }
  
  void *data;
  
private:
  stack *pre, *next;
  
  static stack *main;
};

struct scrollsize {
  int start, end;
};

struct mathnode {
  int next;
  int prev;
  int number;
  struct listnode *var;
  signed char pri;
  signed char opp;
};

class listnode;

struct waitcommdata {
  unsigned char *comm;
  int time;
  struct waitcommdata *next;
};

struct set_args_elem {
  const uchar *name;
  ulong bit;
  int is_local;
  const char *descr;
};

#define MASTER_FLAG_BELL         (1L<<0)    // bell on/off
#define MASTER_FLAG_NO_NEG_OFF   (1L<<1)    // don't use neg off ansi code
#define MASTER_FLAG_VISUAL_BELL  (1L<<2)    // visual belling on/off

class master_table_data {
public:
  master_table_data();
  
  unsigned char *history[HISTORY_SIZE];
  int history_pos;
  unsigned char inputbuf[BUFFER_SIZE];
  int inputbuf_pos, inputbuf2_pos, cursor_pos, utexrem, ucommand;
  int aliascalls;
  window *output_window;
  session *main_session;

  int is_set_flag(const ulong bit) const { return(master_flags & bit); }
  void set_flag(const ulong bit)         { master_flags |= bit; }
  void rem_flag(const ulong bit)         { master_flags &= ~bit; }
  
private:
  ulong master_flags;
};
